#include "cuda_helper.h"
#include <memory.h> // memcpy()
#include "cuda_vector.h"


#define TPB 320

__constant__ uint32_t c_PaddedMessage80[32]; // padded message (80 bytes + padding)

// #ifdef NOASM
#include "cuda_x11_aes_noasm.cu"
// #else
// #include "cuda_x11_aes.cu"
// #endif

static __device__ __forceinline__
void AES_ROUND_NOKEY(
	const uint32_t*const __restrict__ sharedMemory,
	uint32_t &x0, uint32_t &x1, uint32_t &x2, uint32_t &x3)
{
	aes_round(sharedMemory,
			  x0, x1, x2, x3,
			  x0, x1, x2, x3);
}

static __device__ __forceinline__
void KEY_EXPAND_ELT(
	const uint32_t*const __restrict__ sharedMemory,
	uint32_t &k0, uint32_t &k1, uint32_t &k2, uint32_t &k3)
{
	uint32_t y0, y1, y2, y3;
	aes_round(sharedMemory,
			  k0, k1, k2, k3,
			  y0, y1, y2, y3);

	k0 = y1;
	k1 = y2;
	k2 = y3;
	k3 = y0;
}

static __device__ __forceinline__
void shavite_gpu_init(uint32_t *sharedMemory)
{
	/* each thread startup will fill a uint32 */
	if (threadIdx.x < 256) {
		/* each thread startup will fill a uint32 */
		sharedMemory[threadIdx.x] = d_AES0[threadIdx.x];
		sharedMemory[threadIdx.x + 256] = ROL8(sharedMemory[threadIdx.x]);
		sharedMemory[threadIdx.x + 512] = ROL16(sharedMemory[threadIdx.x]);
		sharedMemory[threadIdx.x + 768] = ROL24(sharedMemory[threadIdx.x]);
//		sharedMemory[threadIdx.x + 64 * 2 ] = d_AES0[threadIdx.x + 64 * 2];
//		sharedMemory[threadIdx.x + 64 * 2 + 256] = d_AES1[threadIdx.x + 64 * 2];
//		sharedMemory[threadIdx.x + 64 * 2 + 512] = d_AES2[threadIdx.x + 64 * 2];
//		sharedMemory[threadIdx.x + 64 * 2 + 768] = d_AES3[threadIdx.x + 64 * 2];
	}
}
__global__ __launch_bounds__(TPB, 3)
void x11_shavite512_gpu_hash_64(uint32_t threads, uint32_t *const __restrict__ g_hash)
{
	__shared__  __align__(128) uint32_t sharedMemory[1024];

	shavite_gpu_init(sharedMemory);
	__syncthreads();
	const uint32_t thread = (blockDim.x * blockIdx.x + threadIdx.x);
	if(thread < threads)
	{
		uint32_t *Hash = &g_hash[thread * 16];

		uint32_t rk[32];
		uint32_t msg[16];

		uint28 *phash = (uint28*)Hash;
		uint28 *outpt = (uint28*)msg;
		outpt[0] = phash[0];
		outpt[1] = phash[1];

		uint32_t state[16] =
		{
			0x72FCCDD8, 0x79CA4727, 0x128A077B, 0x40D55AEC,
			0xD1901A06, 0x430AE307, 0xB29F5CD1, 0xDF07FBFC,
			0x8E45D73D, 0x681AB538, 0xBDE86578, 0xDD577E47,
			0xE275EADE, 0x502D9FCD, 0xB9357178, 0x022A4B9A
		};

		uint32_t x0 = 0xD1901A06;
		uint32_t x1 = 0x430AE307;
		uint32_t x2 = 0xB29F5CD1;
		uint32_t x3 = 0xDF07FBFC;

		for(int i = 0; i < 16; i += 4)
		{

			rk[i + 0] = msg[i + 0];
			x0 ^= msg[i + 0];
			rk[i + 1] = msg[i + 1];
			x1 ^= msg[i + 1];
			rk[i + 2] = msg[i + 2];
			x2 ^= msg[i + 2];
			rk[i + 3] = msg[i + 3];
			x3 ^= msg[i + 3];
			AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		}
		state[0] ^= x0;
		state[1] ^= x1;
		state[2] ^= x2;
		state[3] ^= x3;

		// 1
		KEY_EXPAND_ELT(sharedMemory, rk[0], rk[1], rk[2], rk[3]);

		rk[3] ^= (0x02000000UL ^ 0xFFFFFFFFUL);	//rk[31];
		rk[0] ^= 512;
		//	rk[3] ^= 0xFFFFFFFF;

		x0 = state[0] ^ rk[0];
		x1 = state[1] ^ rk[1];
		x2 = state[2] ^ rk[2];
		x3 = state[3] ^ rk[3];


		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[4], rk[5], rk[6], rk[7]);
		rk[4] ^= rk[0];
		rk[5] ^= rk[1];
		rk[6] ^= rk[2];
		rk[7] ^= rk[3];
		x0 ^= rk[4];
		x1 ^= rk[5];
		x2 ^= rk[6];
		x3 ^= rk[7];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[8], rk[9], rk[10], rk[11]);
		rk[8] ^= rk[4];
		rk[9] ^= rk[5];
		rk[10] ^= rk[6];
		rk[11] ^= rk[7];
		x0 ^= rk[8];
		x1 ^= rk[9];
		x2 ^= rk[10];
		x3 ^= rk[11];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[12], rk[13], rk[14], rk[15]);
		rk[12] ^= rk[8];
		rk[13] ^= rk[9];
		rk[14] ^= rk[10];
		rk[15] ^= rk[11];
		x0 ^= rk[12];
		x1 ^= rk[13];
		x2 ^= rk[14];
		x3 ^= rk[15];

		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);

		state[8] ^= 0x32be246fUL;
		state[9] ^= 0xe33ad1e5UL;
		state[10] ^= 0xd659b13eUL;
		state[11] ^= 0xb6a1a92cUL;

		state[12] ^= x0;
		state[13] ^= x1;
		state[14] ^= x2;
		state[15] ^= x3;

		rk[16] = rk[12] ^ 0x63636363UL;
		rk[17] = rk[13] ^ 0x63636363UL;
		rk[18] = rk[14] ^ 0x63636363UL;
		rk[19] = rk[15] ^ 0x8acdcd24UL;
		x0 = state[8] ^ rk[16];
		x1 = state[9] ^ rk[17];
		x2 = state[10] ^ rk[18];
		x3 = state[11] ^ rk[19];
		rk[20] = 0x63636363UL ^ rk[16];
		rk[21] = 0x63636363UL ^ rk[17];
		rk[22] = 0x63636363UL ^ rk[18];
		rk[23] = 0x63636363UL ^ rk[19];
		rk[24] = 0x63636363UL ^ rk[20];
		rk[25] = 0x63636363UL ^ rk[21];
		rk[26] = 0x63636363UL ^ rk[22];
		rk[27] = 0x4b5f7777UL ^ rk[23];

		rk[28] = 0x63636363UL ^ rk[24];
		rk[29] = 0x63636363UL ^ rk[25];
		rk[30] = 0x63636363UL ^ rk[26];
		rk[31] = 0x4b5f7777UL ^ rk[27];



		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);

		x0 ^= rk[20];
		x1 ^= rk[21];
		x2 ^= rk[22];
		x3 ^= rk[23];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);


		x0 ^= rk[24];
		x1 ^= rk[25];
		x2 ^= rk[26];
		x3 ^= rk[27];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);

		x0 ^= rk[28];
		x1 ^= rk[29];
		x2 ^= rk[30];
		x3 ^= rk[31];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);

		state[4] ^= x0;
		state[5] ^= x1;
		state[6] ^= x2;
		state[7] ^= x3;

		rk[0] ^= rk[25];
		x0 = state[12] ^ rk[0];
		rk[1] ^= rk[26];
		x1 = state[13] ^ rk[1];
		rk[2] ^= rk[27];
		x2 = state[14] ^ rk[2];
		rk[3] ^= rk[28];
		x3 = state[15] ^ rk[3];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[4] ^= rk[29];
		x0 ^= rk[4];
		rk[5] ^= rk[30];
		x1 ^= rk[5];
		rk[6] ^= rk[31];
		x2 ^= rk[6];
		rk[7] ^= rk[0];
		x3 ^= rk[7];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[8] ^= rk[1];
		x0 ^= rk[8];
		rk[9] ^= rk[2];
		x1 ^= rk[9];
		rk[10] ^= rk[3];
		x2 ^= rk[10];
		rk[11] ^= rk[4];
		x3 ^= rk[11];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[12] ^= rk[5];
		x0 ^= rk[12];
		rk[13] ^= rk[6];
		x1 ^= rk[13];
		rk[14] ^= rk[7];
		x2 ^= rk[14];
		rk[15] ^= rk[8];
		x3 ^= rk[15];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		state[8] ^= x0;
		state[9] ^= x1;
		state[10] ^= x2;
		state[11] ^= x3;
		rk[16] ^= rk[9];
		x0 = state[4] ^ rk[16];
		rk[17] ^= rk[10];
		x1 = state[5] ^ rk[17];
		rk[18] ^= rk[11];
		x2 = state[6] ^ rk[18];
		rk[19] ^= rk[12];
		x3 = state[7] ^ rk[19];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[20] ^= rk[13];
		x0 ^= rk[20];
		rk[21] ^= rk[14];
		x1 ^= rk[21];
		rk[22] ^= rk[15];
		x2 ^= rk[22];
		rk[23] ^= rk[16];
		x3 ^= rk[23];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[24] ^= rk[17];
		x0 ^= rk[24];
		rk[25] ^= rk[18];
		x1 ^= rk[25];
		rk[26] ^= rk[19];
		x2 ^= rk[26];
		rk[27] ^= rk[20];
		x3 ^= rk[27];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[28] ^= rk[21];
		x0 ^= rk[28];
		rk[29] ^= rk[22];
		x1 ^= rk[29];
		rk[30] ^= rk[23];
		x2 ^= rk[30];
		rk[31] ^= rk[24];
		x3 ^= rk[31];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		state[0] ^= x0;
		state[1] ^= x1;
		state[2] ^= x2;
		state[3] ^= x3;
		/* round 3, 7, 11 */
		KEY_EXPAND_ELT(sharedMemory, rk[0], rk[1], rk[2], rk[3]);
		rk[0] ^= rk[28];
		rk[1] ^= rk[29];
		rk[2] ^= rk[30];
		rk[3] ^= rk[31];
		x0 = state[8] ^ rk[0];
		x1 = state[9] ^ rk[1];
		x2 = state[10] ^ rk[2];
		x3 = state[11] ^ rk[3];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[4], rk[5], rk[6], rk[7]);
		rk[4] ^= rk[0];
		rk[5] ^= rk[1];
		rk[6] ^= rk[2];
		rk[7] ^= rk[3];
		x0 ^= rk[4];
		x1 ^= rk[5];
		x2 ^= rk[6];
		x3 ^= rk[7];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[8], rk[9], rk[10], rk[11]);
		rk[8] ^= rk[4];
		rk[9] ^= rk[5];
		rk[10] ^= rk[6];
		rk[11] ^= rk[7];
		x0 ^= rk[8];
		x1 ^= rk[9];
		x2 ^= rk[10];
		x3 ^= rk[11];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[12], rk[13], rk[14], rk[15]);
		rk[12] ^= rk[8];
		rk[13] ^= rk[9];
		rk[14] ^= rk[10];
		rk[15] ^= rk[11];
		x0 ^= rk[12];
		x1 ^= rk[13];
		x2 ^= rk[14];
		x3 ^= rk[15];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		state[4] ^= x0;
		state[5] ^= x1;
		state[6] ^= x2;
		state[7] ^= x3;
		KEY_EXPAND_ELT(sharedMemory, rk[16], rk[17], rk[18], rk[19]);
		rk[16] ^= rk[12];
		rk[17] ^= rk[13];
		rk[18] ^= rk[14];
		rk[19] ^= rk[15];
		x0 = state[0] ^ rk[16];
		x1 = state[1] ^ rk[17];
		x2 = state[2] ^ rk[18];
		x3 = state[3] ^ rk[19];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[20], rk[21], rk[22], rk[23]);
		rk[20] ^= rk[16];
		rk[21] ^= rk[17];
		rk[22] ^= rk[18];
		rk[23] ^= rk[19];
		x0 ^= rk[20];
		x1 ^= rk[21];
		x2 ^= rk[22];
		x3 ^= rk[23];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[24], rk[25], rk[26], rk[27]);
		rk[24] ^= rk[20];
		rk[25] ^= rk[21];
		rk[26] ^= rk[22];
		rk[27] ^= rk[23];
		x0 ^= rk[24];
		x1 ^= rk[25];
		x2 ^= rk[26];
		x3 ^= rk[27];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[28], rk[29], rk[30], rk[31]);
		rk[28] ^= rk[24];
		rk[29] ^= rk[25];
		rk[30] ^= rk[26];
		rk[31] ^= rk[27];
		x0 ^= rk[28];
		x1 ^= rk[29];
		x2 ^= rk[30];
		x3 ^= rk[31];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		state[12] ^= x0;
		state[13] ^= x1;
		state[14] ^= x2;
		state[15] ^= x3;
		/* round 4, 8, 12 */
		rk[0] ^= rk[25];
		x0 = state[4] ^ rk[0];
		rk[1] ^= rk[26];
		x1 = state[5] ^ rk[1];
		rk[2] ^= rk[27];
		x2 = state[6] ^ rk[2];
		rk[3] ^= rk[28];
		x3 = state[7] ^ rk[3];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[4] ^= rk[29];
		x0 ^= rk[4];
		rk[5] ^= rk[30];
		x1 ^= rk[5];
		rk[6] ^= rk[31];
		x2 ^= rk[6];
		rk[7] ^= rk[0];
		x3 ^= rk[7];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[8] ^= rk[1];
		x0 ^= rk[8];
		rk[9] ^= rk[2];
		x1 ^= rk[9];
		rk[10] ^= rk[3];
		x2 ^= rk[10];
		rk[11] ^= rk[4];
		x3 ^= rk[11];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[12] ^= rk[5];
		x0 ^= rk[12];
		rk[13] ^= rk[6];
		x1 ^= rk[13];
		rk[14] ^= rk[7];
		x2 ^= rk[14];
		rk[15] ^= rk[8];
		x3 ^= rk[15];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);

		state[0] ^= x0;
		state[1] ^= x1;
		state[2] ^= x2;
		state[3] ^= x3;
		rk[16] ^= rk[9];
		x0 = state[12] ^ rk[16];
		rk[17] ^= rk[10];
		x1 = state[13] ^ rk[17];
		rk[18] ^= rk[11];
		x2 = state[14] ^ rk[18];
		rk[19] ^= rk[12];
		x3 = state[15] ^ rk[19];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[20] ^= rk[13];
		x0 ^= rk[20];
		rk[21] ^= rk[14];
		x1 ^= rk[21];
		rk[22] ^= rk[15];
		x2 ^= rk[22];
		rk[23] ^= rk[16];
		x3 ^= rk[23];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[24] ^= rk[17];
		x0 ^= rk[24];
		rk[25] ^= rk[18];
		x1 ^= rk[25];
		rk[26] ^= rk[19];
		x2 ^= rk[26];
		rk[27] ^= rk[20];
		x3 ^= rk[27];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[28] ^= rk[21];
		x0 ^= rk[28];
		rk[29] ^= rk[22];
		x1 ^= rk[29];
		rk[30] ^= rk[23];
		x2 ^= rk[30];
		rk[31] ^= rk[24];
		x3 ^= rk[31];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		state[8] ^= x0;
		state[9] ^= x1;
		state[10] ^= x2;
		state[11] ^= x3;

		// 2
		KEY_EXPAND_ELT(sharedMemory, rk[0], rk[1], rk[2], rk[3]);
		rk[0] ^= rk[28];
		rk[1] ^= rk[29];
		rk[2] ^= rk[30];
		rk[3] ^= rk[31];
		x0 = state[0] ^ rk[0];
		x1 = state[1] ^ rk[1];
		x2 = state[2] ^ rk[2];
		x3 = state[3] ^ rk[3];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[4], rk[5], rk[6], rk[7]);
		rk[4] ^= rk[0];
		rk[5] ^= rk[1];
		rk[6] ^= rk[2];
		rk[7] ^= rk[3];
		rk[7] ^= ~512;
		x0 ^= rk[4];
		x1 ^= rk[5];
		x2 ^= rk[6];
		x3 ^= rk[7];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[8], rk[9], rk[10], rk[11]);
		rk[8] ^= rk[4];
		rk[9] ^= rk[5];
		rk[10] ^= rk[6];
		rk[11] ^= rk[7];
		x0 ^= rk[8];
		x1 ^= rk[9];
		x2 ^= rk[10];
		x3 ^= rk[11];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[12], rk[13], rk[14], rk[15]);
		rk[12] ^= rk[8];
		rk[13] ^= rk[9];
		rk[14] ^= rk[10];
		rk[15] ^= rk[11];
		x0 ^= rk[12];
		x1 ^= rk[13];
		x2 ^= rk[14];
		x3 ^= rk[15];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		state[12] ^= x0;
		state[13] ^= x1;
		state[14] ^= x2;
		state[15] ^= x3;
		KEY_EXPAND_ELT(sharedMemory, rk[16], rk[17], rk[18], rk[19]);
		rk[16] ^= rk[12];
		rk[17] ^= rk[13];
		rk[18] ^= rk[14];
		rk[19] ^= rk[15];
		x0 = state[8] ^ rk[16];
		x1 = state[9] ^ rk[17];
		x2 = state[10] ^ rk[18];
		x3 = state[11] ^ rk[19];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[20], rk[21], rk[22], rk[23]);
		rk[20] ^= rk[16];
		rk[21] ^= rk[17];
		rk[22] ^= rk[18];
		rk[23] ^= rk[19];
		x0 ^= rk[20];
		x1 ^= rk[21];
		x2 ^= rk[22];
		x3 ^= rk[23];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[24], rk[25], rk[26], rk[27]);
		rk[24] ^= rk[20];
		rk[25] ^= rk[21];
		rk[26] ^= rk[22];
		rk[27] ^= rk[23];
		x0 ^= rk[24];
		x1 ^= rk[25];
		x2 ^= rk[26];
		x3 ^= rk[27];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[28], rk[29], rk[30], rk[31]);
		rk[28] ^= rk[24];
		rk[29] ^= rk[25];
		rk[30] ^= rk[26];
		rk[31] ^= rk[27];
		x0 ^= rk[28];
		x1 ^= rk[29];
		x2 ^= rk[30];
		x3 ^= rk[31];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		state[4] ^= x0;
		state[5] ^= x1;
		state[6] ^= x2;
		state[7] ^= x3;

		rk[0] ^= rk[25];
		x0 = state[12] ^ rk[0];
		rk[1] ^= rk[26];
		x1 = state[13] ^ rk[1];
		rk[2] ^= rk[27];
		x2 = state[14] ^ rk[2];
		rk[3] ^= rk[28];
		x3 = state[15] ^ rk[3];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[4] ^= rk[29];
		x0 ^= rk[4];
		rk[5] ^= rk[30];
		x1 ^= rk[5];
		rk[6] ^= rk[31];
		x2 ^= rk[6];
		rk[7] ^= rk[0];
		x3 ^= rk[7];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[8] ^= rk[1];
		x0 ^= rk[8];
		rk[9] ^= rk[2];
		x1 ^= rk[9];
		rk[10] ^= rk[3];
		x2 ^= rk[10];
		rk[11] ^= rk[4];
		x3 ^= rk[11];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[12] ^= rk[5];
		x0 ^= rk[12];
		rk[13] ^= rk[6];
		x1 ^= rk[13];
		rk[14] ^= rk[7];
		x2 ^= rk[14];
		rk[15] ^= rk[8];
		x3 ^= rk[15];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		state[8] ^= x0;
		state[9] ^= x1;
		state[10] ^= x2;
		state[11] ^= x3;
		rk[16] ^= rk[9];
		x0 = state[4] ^ rk[16];
		rk[17] ^= rk[10];
		x1 = state[5] ^ rk[17];
		rk[18] ^= rk[11];
		x2 = state[6] ^ rk[18];
		rk[19] ^= rk[12];
		x3 = state[7] ^ rk[19];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[20] ^= rk[13];
		x0 ^= rk[20];
		rk[21] ^= rk[14];
		x1 ^= rk[21];
		rk[22] ^= rk[15];
		x2 ^= rk[22];
		rk[23] ^= rk[16];
		x3 ^= rk[23];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[24] ^= rk[17];
		x0 ^= rk[24];
		rk[25] ^= rk[18];
		x1 ^= rk[25];
		rk[26] ^= rk[19];
		x2 ^= rk[26];
		rk[27] ^= rk[20];
		x3 ^= rk[27];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[28] ^= rk[21];
		x0 ^= rk[28];
		rk[29] ^= rk[22];
		x1 ^= rk[29];
		rk[30] ^= rk[23];
		x2 ^= rk[30];
		rk[31] ^= rk[24];
		x3 ^= rk[31];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		state[0] ^= x0;
		state[1] ^= x1;
		state[2] ^= x2;
		state[3] ^= x3;
		/* round 3, 7, 11 */
		KEY_EXPAND_ELT(sharedMemory, rk[0], rk[1], rk[2], rk[3]);
		rk[0] ^= rk[28];
		rk[1] ^= rk[29];
		rk[2] ^= rk[30];
		rk[3] ^= rk[31];
		x0 = state[8] ^ rk[0];
		x1 = state[9] ^ rk[1];
		x2 = state[10] ^ rk[2];
		x3 = state[11] ^ rk[3];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[4], rk[5], rk[6], rk[7]);
		rk[4] ^= rk[0];
		rk[5] ^= rk[1];
		rk[6] ^= rk[2];
		rk[7] ^= rk[3];
		x0 ^= rk[4];
		x1 ^= rk[5];
		x2 ^= rk[6];
		x3 ^= rk[7];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[8], rk[9], rk[10], rk[11]);
		rk[8] ^= rk[4];
		rk[9] ^= rk[5];
		rk[10] ^= rk[6];
		rk[11] ^= rk[7];
		x0 ^= rk[8];
		x1 ^= rk[9];
		x2 ^= rk[10];
		x3 ^= rk[11];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[12], rk[13], rk[14], rk[15]);
		rk[12] ^= rk[8];
		rk[13] ^= rk[9];
		rk[14] ^= rk[10];
		rk[15] ^= rk[11];
		x0 ^= rk[12];
		x1 ^= rk[13];
		x2 ^= rk[14];
		x3 ^= rk[15];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		state[4] ^= x0;
		state[5] ^= x1;
		state[6] ^= x2;
		state[7] ^= x3;
		KEY_EXPAND_ELT(sharedMemory, rk[16], rk[17], rk[18], rk[19]);
		rk[16] ^= rk[12];
		rk[17] ^= rk[13];
		rk[18] ^= rk[14];
		rk[19] ^= rk[15];
		x0 = state[0] ^ rk[16];
		x1 = state[1] ^ rk[17];
		x2 = state[2] ^ rk[18];
		x3 = state[3] ^ rk[19];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[20], rk[21], rk[22], rk[23]);
		rk[20] ^= rk[16];
		rk[21] ^= rk[17];
		rk[22] ^= rk[18];
		rk[23] ^= rk[19];
		x0 ^= rk[20];
		x1 ^= rk[21];
		x2 ^= rk[22];
		x3 ^= rk[23];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[24], rk[25], rk[26], rk[27]);
		rk[24] ^= rk[20];
		rk[25] ^= rk[21];
		rk[26] ^= rk[22];
		rk[27] ^= rk[23];
		x0 ^= rk[24];
		x1 ^= rk[25];
		x2 ^= rk[26];
		x3 ^= rk[27];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[28], rk[29], rk[30], rk[31]);
		rk[28] ^= rk[24];
		rk[29] ^= rk[25];
		rk[30] ^= rk[26];
		rk[31] ^= rk[27];
		x0 ^= rk[28];
		x1 ^= rk[29];
		x2 ^= rk[30];
		x3 ^= rk[31];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		state[12] ^= x0;
		state[13] ^= x1;
		state[14] ^= x2;
		state[15] ^= x3;
		/* round 4, 8, 12 */
		rk[0] ^= rk[25];
		x0 = state[4] ^ rk[0];
		rk[1] ^= rk[26];
		x1 = state[5] ^ rk[1];
		rk[2] ^= rk[27];
		x2 = state[6] ^ rk[2];
		rk[3] ^= rk[28];
		x3 = state[7] ^ rk[3];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[4] ^= rk[29];
		x0 ^= rk[4];
		rk[5] ^= rk[30];
		x1 ^= rk[5];
		rk[6] ^= rk[31];
		x2 ^= rk[6];
		rk[7] ^= rk[0];
		x3 ^= rk[7];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[8] ^= rk[1];
		x0 ^= rk[8];
		rk[9] ^= rk[2];
		x1 ^= rk[9];
		rk[10] ^= rk[3];
		x2 ^= rk[10];
		rk[11] ^= rk[4];
		x3 ^= rk[11];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[12] ^= rk[5];
		x0 ^= rk[12];
		rk[13] ^= rk[6];
		x1 ^= rk[13];
		rk[14] ^= rk[7];
		x2 ^= rk[14];
		rk[15] ^= rk[8];
		x3 ^= rk[15];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		state[0] ^= x0;
		state[1] ^= x1;
		state[2] ^= x2;
		state[3] ^= x3;
		rk[16] ^= rk[9];
		x0 = state[12] ^ rk[16];
		rk[17] ^= rk[10];
		x1 = state[13] ^ rk[17];
		rk[18] ^= rk[11];
		x2 = state[14] ^ rk[18];
		rk[19] ^= rk[12];
		x3 = state[15] ^ rk[19];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[20] ^= rk[13];
		x0 ^= rk[20];
		rk[21] ^= rk[14];
		x1 ^= rk[21];
		rk[22] ^= rk[15];
		x2 ^= rk[22];
		rk[23] ^= rk[16];
		x3 ^= rk[23];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[24] ^= rk[17];
		x0 ^= rk[24];
		rk[25] ^= rk[18];
		x1 ^= rk[25];
		rk[26] ^= rk[19];
		x2 ^= rk[26];
		rk[27] ^= rk[20];
		x3 ^= rk[27];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[28] ^= rk[21];
		x0 ^= rk[28];
		rk[29] ^= rk[22];
		x1 ^= rk[29];
		rk[30] ^= rk[23];
		x2 ^= rk[30];
		rk[31] ^= rk[24];
		x3 ^= rk[31];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		state[8] ^= x0;
		state[9] ^= x1;
		state[10] ^= x2;
		state[11] ^= x3;

		// 3
		KEY_EXPAND_ELT(sharedMemory, rk[0], rk[1], rk[2], rk[3]);
		rk[0] ^= rk[28];
		rk[1] ^= rk[29];
		rk[2] ^= rk[30];
		rk[3] ^= rk[31];
		x0 = state[0] ^ rk[0];
		x1 = state[1] ^ rk[1];
		x2 = state[2] ^ rk[2];
		x3 = state[3] ^ rk[3];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[4], rk[5], rk[6], rk[7]);
		rk[4] ^= rk[0];
		rk[5] ^= rk[1];
		rk[6] ^= rk[2];
		rk[7] ^= rk[3];
		x0 ^= rk[4];
		x1 ^= rk[5];
		x2 ^= rk[6];
		x3 ^= rk[7];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[8], rk[9], rk[10], rk[11]);
		rk[8] ^= rk[4];
		rk[9] ^= rk[5];
		rk[10] ^= rk[6];
		rk[11] ^= rk[7];
		x0 ^= rk[8];
		x1 ^= rk[9];
		x2 ^= rk[10];
		x3 ^= rk[11];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[12], rk[13], rk[14], rk[15]);
		rk[12] ^= rk[8];
		rk[13] ^= rk[9];
		rk[14] ^= rk[10];
		rk[15] ^= rk[11];
		x0 ^= rk[12];
		x1 ^= rk[13];
		x2 ^= rk[14];
		x3 ^= rk[15];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		state[12] ^= x0;
		state[13] ^= x1;
		state[14] ^= x2;
		state[15] ^= x3;
		KEY_EXPAND_ELT(sharedMemory, rk[16], rk[17], rk[18], rk[19]);
		rk[16] ^= rk[12];
		rk[17] ^= rk[13];
		rk[18] ^= rk[14];
		rk[19] ^= rk[15];
		x0 = state[8] ^ rk[16];
		x1 = state[9] ^ rk[17];
		x2 = state[10] ^ rk[18];
		x3 = state[11] ^ rk[19];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[20], rk[21], rk[22], rk[23]);
		rk[20] ^= rk[16];
		rk[21] ^= rk[17];
		rk[22] ^= rk[18];
		rk[23] ^= rk[19];
		x0 ^= rk[20];
		x1 ^= rk[21];
		x2 ^= rk[22];
		x3 ^= rk[23];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[24], rk[25], rk[26], rk[27]);
		rk[24] ^= rk[20];
		rk[25] ^= rk[21];
		rk[26] ^= rk[22];
		rk[27] ^= rk[23];
		x0 ^= rk[24];
		x1 ^= rk[25];
		x2 ^= rk[26];
		x3 ^= rk[27];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[28], rk[29], rk[30], rk[31]);
		rk[28] ^= rk[24];
		rk[29] ^= rk[25];
		rk[30] ^= rk[26];
		rk[31] ^= ~rk[27];
		rk[30] ^= 512;
		//		rk[31] ^= 0xFFFFFFFF;
		x0 ^= rk[28];
		x1 ^= rk[29];
		x2 ^= rk[30];
		x3 ^= rk[31];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		state[4] ^= x0;
		state[5] ^= x1;
		state[6] ^= x2;
		state[7] ^= x3;

		rk[0] ^= rk[25];
		x0 = state[12] ^ rk[0];
		rk[1] ^= rk[26];
		x1 = state[13] ^ rk[1];
		rk[2] ^= rk[27];
		x2 = state[14] ^ rk[2];
		rk[3] ^= rk[28];
		x3 = state[15] ^ rk[3];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[4] ^= rk[29];
		x0 ^= rk[4];
		rk[5] ^= rk[30];
		x1 ^= rk[5];
		rk[6] ^= rk[31];
		x2 ^= rk[6];
		rk[7] ^= rk[0];
		x3 ^= rk[7];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[8] ^= rk[1];
		x0 ^= rk[8];
		rk[9] ^= rk[2];
		x1 ^= rk[9];
		rk[10] ^= rk[3];
		x2 ^= rk[10];
		rk[11] ^= rk[4];
		x3 ^= rk[11];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[12] ^= rk[5];
		x0 ^= rk[12];
		rk[13] ^= rk[6];
		x1 ^= rk[13];
		rk[14] ^= rk[7];
		x2 ^= rk[14];
		rk[15] ^= rk[8];
		x3 ^= rk[15];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		state[8] ^= x0;
		state[9] ^= x1;
		state[10] ^= x2;
		state[11] ^= x3;
		rk[16] ^= rk[9];
		x0 = state[4] ^ rk[16];
		rk[17] ^= rk[10];
		x1 = state[5] ^ rk[17];
		rk[18] ^= rk[11];
		x2 = state[6] ^ rk[18];
		rk[19] ^= rk[12];
		x3 = state[7] ^ rk[19];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[20] ^= rk[13];
		x0 ^= rk[20];
		rk[21] ^= rk[14];
		x1 ^= rk[21];
		rk[22] ^= rk[15];
		x2 ^= rk[22];
		rk[23] ^= rk[16];
		x3 ^= rk[23];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[24] ^= rk[17];
		x0 ^= rk[24];
		rk[25] ^= rk[18];
		x1 ^= rk[25];
		rk[26] ^= rk[19];
		x2 ^= rk[26];
		rk[27] ^= rk[20];
		x3 ^= rk[27];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[28] ^= rk[21];
		x0 ^= rk[28];
		rk[29] ^= rk[22];
		x1 ^= rk[29];
		rk[30] ^= rk[23];
		x2 ^= rk[30];
		rk[31] ^= rk[24];
		x3 ^= rk[31];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		state[0] ^= x0;
		state[1] ^= x1;
		state[2] ^= x2;
		state[3] ^= x3;

		/* round 3, 7, 11 */
		KEY_EXPAND_ELT(sharedMemory, rk[0], rk[1], rk[2], rk[3]);
		rk[0] ^= rk[28];
		rk[1] ^= rk[29];
		rk[2] ^= rk[30];
		rk[3] ^= rk[31];
		x0 = state[8] ^ rk[0];
		x1 = state[9] ^ rk[1];
		x2 = state[10] ^ rk[2];
		x3 = state[11] ^ rk[3];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[4], rk[5], rk[6], rk[7]);
		rk[4] ^= rk[0];
		rk[5] ^= rk[1];
		rk[6] ^= rk[2];
		rk[7] ^= rk[3];
		x0 ^= rk[4];
		x1 ^= rk[5];
		x2 ^= rk[6];
		x3 ^= rk[7];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[8], rk[9], rk[10], rk[11]);
		rk[8] ^= rk[4];
		rk[9] ^= rk[5];
		rk[10] ^= rk[6];
		rk[11] ^= rk[7];
		x0 ^= rk[8];
		x1 ^= rk[9];
		x2 ^= rk[10];
		x3 ^= rk[11];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[12], rk[13], rk[14], rk[15]);
		rk[12] ^= rk[8];
		rk[13] ^= rk[9];
		rk[14] ^= rk[10];
		rk[15] ^= rk[11];
		x0 ^= rk[12];
		x1 ^= rk[13];
		x2 ^= rk[14];
		x3 ^= rk[15];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		state[4] ^= x0;
		state[5] ^= x1;
		state[6] ^= x2;
		state[7] ^= x3;
		KEY_EXPAND_ELT(sharedMemory, rk[16], rk[17], rk[18], rk[19]);
		rk[16] ^= rk[12];
		rk[17] ^= rk[13];
		rk[18] ^= rk[14];
		rk[19] ^= rk[15];
		x0 = state[0] ^ rk[16];
		x1 = state[1] ^ rk[17];
		x2 = state[2] ^ rk[18];
		x3 = state[3] ^ rk[19];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[20], rk[21], rk[22], rk[23]);
		rk[20] ^= rk[16];
		rk[21] ^= rk[17];
		rk[22] ^= rk[18];
		rk[23] ^= rk[19];
		x0 ^= rk[20];
		x1 ^= rk[21];
		x2 ^= rk[22];
		x3 ^= rk[23];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[24], rk[25], rk[26], rk[27]);
		rk[24] ^= rk[20];
		rk[25] ^= rk[21];
		rk[26] ^= rk[22];
		rk[27] ^= rk[23];
		x0 ^= rk[24];
		x1 ^= rk[25];
		x2 ^= rk[26];
		x3 ^= rk[27];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[28], rk[29], rk[30], rk[31]);
		rk[28] ^= rk[24];
		rk[29] ^= rk[25];
		rk[30] ^= rk[26];
		rk[31] ^= rk[27];
		x0 ^= rk[28];
		x1 ^= rk[29];
		x2 ^= rk[30];
		x3 ^= rk[31];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		state[12] ^= x0;
		state[13] ^= x1;
		state[14] ^= x2;
		state[15] ^= x3;
		/* round 4, 8, 12 */
		rk[0] ^= rk[25];
		x0 = state[4] ^ rk[0];
		rk[1] ^= rk[26];
		x1 = state[5] ^ rk[1];
		rk[2] ^= rk[27];
		x2 = state[6] ^ rk[2];
		rk[3] ^= rk[28];
		x3 = state[7] ^ rk[3];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[4] ^= rk[29];
		x0 ^= rk[4];
		rk[5] ^= rk[30];
		x1 ^= rk[5];
		rk[6] ^= rk[31];
		x2 ^= rk[6];
		rk[7] ^= rk[0];
		x3 ^= rk[7];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[8] ^= rk[1];
		x0 ^= rk[8];
		rk[9] ^= rk[2];
		x1 ^= rk[9];
		rk[10] ^= rk[3];
		x2 ^= rk[10];
		rk[11] ^= rk[4];
		x3 ^= rk[11];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[12] ^= rk[5];
		x0 ^= rk[12];
		rk[13] ^= rk[6];
		x1 ^= rk[13];
		rk[14] ^= rk[7];
		x2 ^= rk[14];
		rk[15] ^= rk[8];
		x3 ^= rk[15];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		state[0] ^= x0;
		state[1] ^= x1;
		state[2] ^= x2;
		state[3] ^= x3;
		rk[16] ^= rk[9];
		x0 = state[12] ^ rk[16];
		rk[17] ^= rk[10];
		x1 = state[13] ^ rk[17];
		rk[18] ^= rk[11];
		x2 = state[14] ^ rk[18];
		rk[19] ^= rk[12];
		x3 = state[15] ^ rk[19];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[20] ^= rk[13];
		x0 ^= rk[20];
		rk[21] ^= rk[14];
		x1 ^= rk[21];
		rk[22] ^= rk[15];
		x2 ^= rk[22];
		rk[23] ^= rk[16];
		x3 ^= rk[23];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[24] ^= rk[17];
		x0 ^= rk[24];
		rk[25] ^= rk[18];
		x1 ^= rk[25];
		rk[26] ^= rk[19];
		x2 ^= rk[26];
		rk[27] ^= rk[20];
		x3 ^= rk[27];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		rk[28] ^= rk[21];
		x0 ^= rk[28];
		rk[29] ^= rk[22];
		x1 ^= rk[29];
		rk[30] ^= rk[23];
		x2 ^= rk[30];
		rk[31] ^= rk[24];
		x3 ^= rk[31];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		state[8] ^= x0;
		state[9] ^= x1;
		state[10] ^= x2;
		state[11] ^= x3;

		/* round 13 */
		KEY_EXPAND_ELT(sharedMemory, rk[0], rk[1], rk[2], rk[3]);
		rk[0] ^= rk[28];
		rk[1] ^= rk[29];
		rk[2] ^= rk[30];
		rk[3] ^= rk[31];
		x0 = state[0] ^ rk[0];
		x1 = state[1] ^ rk[1];
		x2 = state[2] ^ rk[2];
		x3 = state[3] ^ rk[3];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[4], rk[5], rk[6], rk[7]);
		rk[4] ^= rk[0];
		rk[5] ^= rk[1];
		rk[6] ^= rk[2];
		rk[7] ^= rk[3];
		x0 ^= rk[4];
		x1 ^= rk[5];
		x2 ^= rk[6];
		x3 ^= rk[7];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[8], rk[9], rk[10], rk[11]);
		rk[8] ^= rk[4];
		rk[9] ^= rk[5];
		rk[10] ^= rk[6];
		rk[11] ^= rk[7];
		x0 ^= rk[8];
		x1 ^= rk[9];
		x2 ^= rk[10];
		x3 ^= rk[11];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[12], rk[13], rk[14], rk[15]);
		rk[12] ^= rk[8];
		rk[13] ^= rk[9];
		rk[14] ^= rk[10];
		rk[15] ^= rk[11];
		x0 ^= rk[12];
		x1 ^= rk[13];
		x2 ^= rk[14];
		x3 ^= rk[15];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		state[12] ^= x0;
		state[13] ^= x1;
		state[14] ^= x2;
		state[15] ^= x3;
		KEY_EXPAND_ELT(sharedMemory, rk[16], rk[17], rk[18], rk[19]);
		rk[16] ^= rk[12];
		rk[17] ^= rk[13];
		rk[18] ^= rk[14];
		rk[19] ^= rk[15];
		x0 = state[8] ^ rk[16];
		x1 = state[9] ^ rk[17];
		x2 = state[10] ^ rk[18];
		x3 = state[11] ^ rk[19];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[20], rk[21], rk[22], rk[23]);
		rk[20] ^= rk[16];
		rk[21] ^= rk[17];
		rk[22] ^= rk[18];
		rk[23] ^= rk[19];
		x0 ^= rk[20];
		x1 ^= rk[21];
		x2 ^= rk[22];
		x3 ^= rk[23];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[24], rk[25], rk[26], rk[27]);
		rk[24] ^= rk[20];
		rk[25] ^= rk[21] ^ 512;
		rk[26] ^= rk[22];
		rk[27] ^= ~rk[23]; //^ 0xFFFFFFFF;
		x0 ^= rk[24];
		x1 ^= rk[25];
		x2 ^= rk[26];
		x3 ^= rk[27];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		KEY_EXPAND_ELT(sharedMemory, rk[28], rk[29], rk[30], rk[31]);
		rk[28] ^= rk[24];
		rk[29] ^= rk[25];
		rk[30] ^= rk[26];
		rk[31] ^= rk[27];
		x0 ^= rk[28];
		x1 ^= rk[29];
		x2 ^= rk[30];
		x3 ^= rk[31];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
		state[4] ^= x0;
		state[5] ^= x1;
		state[6] ^= x2;
		state[7] ^= x3;

		Hash[0] = 0x72FCCDD8 ^ state[8];
		Hash[1] = 0x79CA4727 ^ state[9];
		Hash[2] = 0x128A077B ^ state[10];
		Hash[3] = 0x40D55AEC ^ state[11];
		Hash[4] = 0xD1901A06 ^ state[12];
		Hash[5] = 0x430AE307 ^ state[13];
		Hash[6] = 0xB29F5CD1 ^ state[14];
		Hash[7] = 0xDF07FBFC ^ state[15];
		Hash[8] = 0x8E45D73D ^ state[0];
		Hash[9] = 0x681AB538 ^ state[1];
		Hash[10] = 0xBDE86578 ^ state[2];
		Hash[11] = 0xDD577E47 ^ state[3];
		Hash[12] = 0xE275EADE ^ state[4];
		Hash[13] = 0x502D9FCD ^ state[5];
		Hash[14] = 0xB9357178 ^ state[6];
		Hash[15] = 0x022A4B9A ^ state[7];
	}
}

__device__ __forceinline__
static void c512(const uint32_t*const __restrict__ sharedMemory, uint32_t *const __restrict__  state, uint32_t *const __restrict__  msg)
{
	uint32_t p0, p1, p2, p3, p4, p5, p6, p7;
	uint32_t p8, p9, pA, pB, pC, pD, pE, pF;
	uint32_t x0, x1, x2, x3;
	uint32_t rk[32];
	uint32_t i;
	const uint32_t counter = 640;

	p0 = state[0x0];
	p1 = state[0x1];
	p2 = state[0x2];
	p3 = state[0x3];
	p4 = state[0x4];
	p5 = state[0x5];
	p6 = state[0x6];
	p7 = state[0x7];
	p8 = state[0x8];
	p9 = state[0x9];
	pA = state[0xA];
	pB = state[0xB];
	pC = state[0xC];
	pD = state[0xD];
	pE = state[0xE];
	pF = state[0xF];

	x0 = p4;
	x1 = p5;
	x2 = p6;
	x3 = p7;
#pragma unroll
	for(i = 0; i<16; i += 4)
	{
		rk[i] = msg[i];
		x0 ^= msg[i];
		rk[i + 1] = msg[i + 1];
		x1 ^= msg[i + 1];
		rk[i + 2] = msg[i + 2];
		x2 ^= msg[i + 2];
		rk[i + 3] = msg[i + 3];
		x3 ^= msg[i + 3];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	}

	p0 ^= x0;
	p1 ^= x1;
	p2 ^= x2;
	p3 ^= x3;
	x0 = pC;
	x1 = pD;
	x2 = pE;
	x3 = pF;

#pragma unroll 
	for(i = 16; i<32; i += 4)
	{
		rk[i] = msg[i];
		x0 ^= msg[i];
		rk[i + 1] = msg[i + 1];
		x1 ^= msg[i + 1];
		rk[i + 2] = msg[i + 2];
		x2 ^= msg[i + 2];
		rk[i + 3] = msg[i + 3];
		x3 ^= msg[i + 3];
		AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	}
	p8 ^= x0;
	p9 ^= x1;
	pA ^= x2;
	pB ^= x3;

	// 1
	KEY_EXPAND_ELT(sharedMemory, rk[0], rk[1], rk[2], rk[3]);

	rk[0] ^= rk[28];
	rk[1] ^= rk[29];
	rk[2] ^= rk[30];
	rk[3] ^= ~rk[31];
	rk[0] ^= counter;
	//rk[3] ^= 0xFFFFFFFF;
	x0 = p0 ^ rk[0];
	x1 = p1 ^ rk[1];
	x2 = p2 ^ rk[2];
	x3 = p3 ^ rk[3];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[4], rk[5], rk[6], rk[7]);
	rk[4] ^= rk[0];
	rk[5] ^= rk[1];
	rk[6] ^= rk[2];
	rk[7] ^= rk[3];
	x0 ^= rk[4];
	x1 ^= rk[5];
	x2 ^= rk[6];
	x3 ^= rk[7];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[8], rk[9], rk[10], rk[11]);
	rk[8] ^= rk[4];
	rk[9] ^= rk[5];
	rk[10] ^= rk[6];
	rk[11] ^= rk[7];
	x0 ^= rk[8];
	x1 ^= rk[9];
	x2 ^= rk[10];
	x3 ^= rk[11];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[12], rk[13], rk[14], rk[15]);
	rk[12] ^= rk[8];
	rk[13] ^= rk[9];
	rk[14] ^= rk[10];
	rk[15] ^= rk[11];
	x0 ^= rk[12];
	x1 ^= rk[13];
	x2 ^= rk[14];
	x3 ^= rk[15];

	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	pC ^= x0;
	pD ^= x1;
	pE ^= x2;
	pF ^= x3;

	KEY_EXPAND_ELT(sharedMemory, rk[16], rk[17], rk[18], rk[19]);
	rk[16] ^= rk[12];
	rk[17] ^= rk[13];
	rk[18] ^= rk[14];
	rk[19] ^= rk[15];
	x0 = p8 ^ rk[16];
	x1 = p9 ^ rk[17];
	x2 = pA ^ rk[18];
	x3 = pB ^ rk[19];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[20], rk[21], rk[22], rk[23]);
	rk[20] ^= rk[16];
	rk[21] ^= rk[17];
	rk[22] ^= rk[18];
	rk[23] ^= rk[19];
	x0 ^= rk[20];
	x1 ^= rk[21];
	x2 ^= rk[22];
	x3 ^= rk[23];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[24], rk[25], rk[26], rk[27]);
	rk[24] ^= rk[20];
	rk[25] ^= rk[21];
	rk[26] ^= rk[22];
	rk[27] ^= rk[23];
	x0 ^= rk[24];
	x1 ^= rk[25];
	x2 ^= rk[26];
	x3 ^= rk[27];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[28], rk[29], rk[30], rk[31]);
	rk[28] ^= rk[24];
	rk[29] ^= rk[25];
	rk[30] ^= rk[26];
	rk[31] ^= rk[27];
	x0 ^= rk[28];
	x1 ^= rk[29];
	x2 ^= rk[30];
	x3 ^= rk[31];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	p4 ^= x0;
	p5 ^= x1;
	p6 ^= x2;
	p7 ^= x3;

	rk[0] ^= rk[25];
	x0 = pC ^ rk[0];
	rk[1] ^= rk[26];
	x1 = pD ^ rk[1];
	rk[2] ^= rk[27];
	x2 = pE ^ rk[2];
	rk[3] ^= rk[28];
	x3 = pF ^ rk[3];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[4] ^= rk[29];
	x0 ^= rk[4];
	rk[5] ^= rk[30];
	x1 ^= rk[5];
	rk[6] ^= rk[31];
	x2 ^= rk[6];
	rk[7] ^= rk[0];
	x3 ^= rk[7];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[8] ^= rk[1];
	x0 ^= rk[8];
	rk[9] ^= rk[2];
	x1 ^= rk[9];
	rk[10] ^= rk[3];
	x2 ^= rk[10];
	rk[11] ^= rk[4];
	x3 ^= rk[11];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[12] ^= rk[5];
	x0 ^= rk[12];
	rk[13] ^= rk[6];
	x1 ^= rk[13];
	rk[14] ^= rk[7];
	x2 ^= rk[14];
	rk[15] ^= rk[8];
	x3 ^= rk[15];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	p8 ^= x0;
	p9 ^= x1;
	pA ^= x2;
	pB ^= x3;
	rk[16] ^= rk[9];
	x0 = p4 ^ rk[16];
	rk[17] ^= rk[10];
	x1 = p5 ^ rk[17];
	rk[18] ^= rk[11];
	x2 = p6 ^ rk[18];
	rk[19] ^= rk[12];
	x3 = p7 ^ rk[19];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[20] ^= rk[13];
	x0 ^= rk[20];
	rk[21] ^= rk[14];
	x1 ^= rk[21];
	rk[22] ^= rk[15];
	x2 ^= rk[22];
	rk[23] ^= rk[16];
	x3 ^= rk[23];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[24] ^= rk[17];
	x0 ^= rk[24];
	rk[25] ^= rk[18];
	x1 ^= rk[25];
	rk[26] ^= rk[19];
	x2 ^= rk[26];
	rk[27] ^= rk[20];
	x3 ^= rk[27];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[28] ^= rk[21];
	x0 ^= rk[28];
	rk[29] ^= rk[22];
	x1 ^= rk[29];
	rk[30] ^= rk[23];
	x2 ^= rk[30];
	rk[31] ^= rk[24];
	x3 ^= rk[31];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	p0 ^= x0;
	p1 ^= x1;
	p2 ^= x2;
	p3 ^= x3;
	/* round 3, 7, 11 */
	KEY_EXPAND_ELT(sharedMemory, rk[0], rk[1], rk[2], rk[3]);
	rk[0] ^= rk[28];
	rk[1] ^= rk[29];
	rk[2] ^= rk[30];
	rk[3] ^= rk[31];
	x0 = p8 ^ rk[0];
	x1 = p9 ^ rk[1];
	x2 = pA ^ rk[2];
	x3 = pB ^ rk[3];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[4], rk[5], rk[6], rk[7]);
	rk[4] ^= rk[0];
	rk[5] ^= rk[1];
	rk[6] ^= rk[2];
	rk[7] ^= rk[3];
	x0 ^= rk[4];
	x1 ^= rk[5];
	x2 ^= rk[6];
	x3 ^= rk[7];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[8], rk[9], rk[10], rk[11]);
	rk[8] ^= rk[4];
	rk[9] ^= rk[5];
	rk[10] ^= rk[6];
	rk[11] ^= rk[7];
	x0 ^= rk[8];
	x1 ^= rk[9];
	x2 ^= rk[10];
	x3 ^= rk[11];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[12], rk[13], rk[14], rk[15]);
	rk[12] ^= rk[8];
	rk[13] ^= rk[9];
	rk[14] ^= rk[10];
	rk[15] ^= rk[11];
	x0 ^= rk[12];
	x1 ^= rk[13];
	x2 ^= rk[14];
	x3 ^= rk[15];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	p4 ^= x0;
	p5 ^= x1;
	p6 ^= x2;
	p7 ^= x3;
	KEY_EXPAND_ELT(sharedMemory, rk[16], rk[17], rk[18], rk[19]);
	rk[16] ^= rk[12];
	rk[17] ^= rk[13];
	rk[18] ^= rk[14];
	rk[19] ^= rk[15];
	x0 = p0 ^ rk[16];
	x1 = p1 ^ rk[17];
	x2 = p2 ^ rk[18];
	x3 = p3 ^ rk[19];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[20], rk[21], rk[22], rk[23]);
	rk[20] ^= rk[16];
	rk[21] ^= rk[17];
	rk[22] ^= rk[18];
	rk[23] ^= rk[19];
	x0 ^= rk[20];
	x1 ^= rk[21];
	x2 ^= rk[22];
	x3 ^= rk[23];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[24], rk[25], rk[26], rk[27]);
	rk[24] ^= rk[20];
	rk[25] ^= rk[21];
	rk[26] ^= rk[22];
	rk[27] ^= rk[23];
	x0 ^= rk[24];
	x1 ^= rk[25];
	x2 ^= rk[26];
	x3 ^= rk[27];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[28], rk[29], rk[30], rk[31]);
	rk[28] ^= rk[24];
	rk[29] ^= rk[25];
	rk[30] ^= rk[26];
	rk[31] ^= rk[27];
	x0 ^= rk[28];
	x1 ^= rk[29];
	x2 ^= rk[30];
	x3 ^= rk[31];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	pC ^= x0;
	pD ^= x1;
	pE ^= x2;
	pF ^= x3;
	/* round 4, 8, 12 */
	rk[0] ^= rk[25];
	x0 = p4 ^ rk[0];
	rk[1] ^= rk[26];
	x1 = p5 ^ rk[1];
	rk[2] ^= rk[27];
	x2 = p6 ^ rk[2];
	rk[3] ^= rk[28];
	x3 = p7 ^ rk[3];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[4] ^= rk[29];
	x0 ^= rk[4];
	rk[5] ^= rk[30];
	x1 ^= rk[5];
	rk[6] ^= rk[31];
	x2 ^= rk[6];
	rk[7] ^= rk[0];
	x3 ^= rk[7];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[8] ^= rk[1];
	x0 ^= rk[8];
	rk[9] ^= rk[2];
	x1 ^= rk[9];
	rk[10] ^= rk[3];
	x2 ^= rk[10];
	rk[11] ^= rk[4];
	x3 ^= rk[11];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[12] ^= rk[5];
	x0 ^= rk[12];
	rk[13] ^= rk[6];
	x1 ^= rk[13];
	rk[14] ^= rk[7];
	x2 ^= rk[14];
	rk[15] ^= rk[8];
	x3 ^= rk[15];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);

	p0 ^= x0;
	p1 ^= x1;
	p2 ^= x2;
	p3 ^= x3;
	rk[16] ^= rk[9];
	x0 = pC ^ rk[16];
	rk[17] ^= rk[10];
	x1 = pD ^ rk[17];
	rk[18] ^= rk[11];
	x2 = pE ^ rk[18];
	rk[19] ^= rk[12];
	x3 = pF ^ rk[19];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[20] ^= rk[13];
	x0 ^= rk[20];
	rk[21] ^= rk[14];
	x1 ^= rk[21];
	rk[22] ^= rk[15];
	x2 ^= rk[22];
	rk[23] ^= rk[16];
	x3 ^= rk[23];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[24] ^= rk[17];
	x0 ^= rk[24];
	rk[25] ^= rk[18];
	x1 ^= rk[25];
	rk[26] ^= rk[19];
	x2 ^= rk[26];
	rk[27] ^= rk[20];
	x3 ^= rk[27];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[28] ^= rk[21];
	x0 ^= rk[28];
	rk[29] ^= rk[22];
	x1 ^= rk[29];
	rk[30] ^= rk[23];
	x2 ^= rk[30];
	rk[31] ^= rk[24];
	x3 ^= rk[31];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	p8 ^= x0;
	p9 ^= x1;
	pA ^= x2;
	pB ^= x3;

	// 2
	KEY_EXPAND_ELT(sharedMemory, rk[0], rk[1], rk[2], rk[3]);
	rk[0] ^= rk[28];
	rk[1] ^= rk[29];
	rk[2] ^= rk[30];
	rk[3] ^= rk[31];
	x0 = p0 ^ rk[0];
	x1 = p1 ^ rk[1];
	x2 = p2 ^ rk[2];
	x3 = p3 ^ rk[3];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[4], rk[5], rk[6], rk[7]);
	rk[4] ^= rk[0];
	rk[5] ^= rk[1];
	rk[6] ^= rk[2];
	rk[7] ^= rk[3];
	rk[7] ^= ~counter;
	x0 ^= rk[4];
	x1 ^= rk[5];
	x2 ^= rk[6];
	x3 ^= rk[7];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[8], rk[9], rk[10], rk[11]);
	rk[8] ^= rk[4];
	rk[9] ^= rk[5];
	rk[10] ^= rk[6];
	rk[11] ^= rk[7];
	x0 ^= rk[8];
	x1 ^= rk[9];
	x2 ^= rk[10];
	x3 ^= rk[11];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[12], rk[13], rk[14], rk[15]);
	rk[12] ^= rk[8];
	rk[13] ^= rk[9];
	rk[14] ^= rk[10];
	rk[15] ^= rk[11];
	x0 ^= rk[12];
	x1 ^= rk[13];
	x2 ^= rk[14];
	x3 ^= rk[15];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	pC ^= x0;
	pD ^= x1;
	pE ^= x2;
	pF ^= x3;
	KEY_EXPAND_ELT(sharedMemory, rk[16], rk[17], rk[18], rk[19]);
	rk[16] ^= rk[12];
	rk[17] ^= rk[13];
	rk[18] ^= rk[14];
	rk[19] ^= rk[15];
	x0 = p8 ^ rk[16];
	x1 = p9 ^ rk[17];
	x2 = pA ^ rk[18];
	x3 = pB ^ rk[19];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[20], rk[21], rk[22], rk[23]);
	rk[20] ^= rk[16];
	rk[21] ^= rk[17];
	rk[22] ^= rk[18];
	rk[23] ^= rk[19];
	x0 ^= rk[20];
	x1 ^= rk[21];
	x2 ^= rk[22];
	x3 ^= rk[23];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[24], rk[25], rk[26], rk[27]);
	rk[24] ^= rk[20];
	rk[25] ^= rk[21];
	rk[26] ^= rk[22];
	rk[27] ^= rk[23];
	x0 ^= rk[24];
	x1 ^= rk[25];
	x2 ^= rk[26];
	x3 ^= rk[27];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[28], rk[29], rk[30], rk[31]);
	rk[28] ^= rk[24];
	rk[29] ^= rk[25];
	rk[30] ^= rk[26];
	rk[31] ^= rk[27];
	x0 ^= rk[28];
	x1 ^= rk[29];
	x2 ^= rk[30];
	x3 ^= rk[31];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	p4 ^= x0;
	p5 ^= x1;
	p6 ^= x2;
	p7 ^= x3;

	rk[0] ^= rk[25];
	x0 = pC ^ rk[0];
	rk[1] ^= rk[26];
	x1 = pD ^ rk[1];
	rk[2] ^= rk[27];
	x2 = pE ^ rk[2];
	rk[3] ^= rk[28];
	x3 = pF ^ rk[3];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[4] ^= rk[29];
	x0 ^= rk[4];
	rk[5] ^= rk[30];
	x1 ^= rk[5];
	rk[6] ^= rk[31];
	x2 ^= rk[6];
	rk[7] ^= rk[0];
	x3 ^= rk[7];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[8] ^= rk[1];
	x0 ^= rk[8];
	rk[9] ^= rk[2];
	x1 ^= rk[9];
	rk[10] ^= rk[3];
	x2 ^= rk[10];
	rk[11] ^= rk[4];
	x3 ^= rk[11];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[12] ^= rk[5];
	x0 ^= rk[12];
	rk[13] ^= rk[6];
	x1 ^= rk[13];
	rk[14] ^= rk[7];
	x2 ^= rk[14];
	rk[15] ^= rk[8];
	x3 ^= rk[15];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	p8 ^= x0;
	p9 ^= x1;
	pA ^= x2;
	pB ^= x3;
	rk[16] ^= rk[9];
	x0 = p4 ^ rk[16];
	rk[17] ^= rk[10];
	x1 = p5 ^ rk[17];
	rk[18] ^= rk[11];
	x2 = p6 ^ rk[18];
	rk[19] ^= rk[12];
	x3 = p7 ^ rk[19];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[20] ^= rk[13];
	x0 ^= rk[20];
	rk[21] ^= rk[14];
	x1 ^= rk[21];
	rk[22] ^= rk[15];
	x2 ^= rk[22];
	rk[23] ^= rk[16];
	x3 ^= rk[23];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[24] ^= rk[17];
	x0 ^= rk[24];
	rk[25] ^= rk[18];
	x1 ^= rk[25];
	rk[26] ^= rk[19];
	x2 ^= rk[26];
	rk[27] ^= rk[20];
	x3 ^= rk[27];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[28] ^= rk[21];
	x0 ^= rk[28];
	rk[29] ^= rk[22];
	x1 ^= rk[29];
	rk[30] ^= rk[23];
	x2 ^= rk[30];
	rk[31] ^= rk[24];
	x3 ^= rk[31];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	p0 ^= x0;
	p1 ^= x1;
	p2 ^= x2;
	p3 ^= x3;
	/* round 3, 7, 11 */
	KEY_EXPAND_ELT(sharedMemory, rk[0], rk[1], rk[2], rk[3]);
	rk[0] ^= rk[28];
	rk[1] ^= rk[29];
	rk[2] ^= rk[30];
	rk[3] ^= rk[31];
	x0 = p8 ^ rk[0];
	x1 = p9 ^ rk[1];
	x2 = pA ^ rk[2];
	x3 = pB ^ rk[3];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[4], rk[5], rk[6], rk[7]);
	rk[4] ^= rk[0];
	rk[5] ^= rk[1];
	rk[6] ^= rk[2];
	rk[7] ^= rk[3];
	x0 ^= rk[4];
	x1 ^= rk[5];
	x2 ^= rk[6];
	x3 ^= rk[7];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[8], rk[9], rk[10], rk[11]);
	rk[8] ^= rk[4];
	rk[9] ^= rk[5];
	rk[10] ^= rk[6];
	rk[11] ^= rk[7];
	x0 ^= rk[8];
	x1 ^= rk[9];
	x2 ^= rk[10];
	x3 ^= rk[11];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[12], rk[13], rk[14], rk[15]);
	rk[12] ^= rk[8];
	rk[13] ^= rk[9];
	rk[14] ^= rk[10];
	rk[15] ^= rk[11];
	x0 ^= rk[12];
	x1 ^= rk[13];
	x2 ^= rk[14];
	x3 ^= rk[15];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	p4 ^= x0;
	p5 ^= x1;
	p6 ^= x2;
	p7 ^= x3;
	KEY_EXPAND_ELT(sharedMemory, rk[16], rk[17], rk[18], rk[19]);
	rk[16] ^= rk[12];
	rk[17] ^= rk[13];
	rk[18] ^= rk[14];
	rk[19] ^= rk[15];
	x0 = p0 ^ rk[16];
	x1 = p1 ^ rk[17];
	x2 = p2 ^ rk[18];
	x3 = p3 ^ rk[19];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[20], rk[21], rk[22], rk[23]);
	rk[20] ^= rk[16];
	rk[21] ^= rk[17];
	rk[22] ^= rk[18];
	rk[23] ^= rk[19];
	x0 ^= rk[20];
	x1 ^= rk[21];
	x2 ^= rk[22];
	x3 ^= rk[23];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[24], rk[25], rk[26], rk[27]);
	rk[24] ^= rk[20];
	rk[25] ^= rk[21];
	rk[26] ^= rk[22];
	rk[27] ^= rk[23];
	x0 ^= rk[24];
	x1 ^= rk[25];
	x2 ^= rk[26];
	x3 ^= rk[27];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[28], rk[29], rk[30], rk[31]);
	rk[28] ^= rk[24];
	rk[29] ^= rk[25];
	rk[30] ^= rk[26];
	rk[31] ^= rk[27];
	x0 ^= rk[28];
	x1 ^= rk[29];
	x2 ^= rk[30];
	x3 ^= rk[31];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	pC ^= x0;
	pD ^= x1;
	pE ^= x2;
	pF ^= x3;
	/* round 4, 8, 12 */
	rk[0] ^= rk[25];
	x0 = p4 ^ rk[0];
	rk[1] ^= rk[26];
	x1 = p5 ^ rk[1];
	rk[2] ^= rk[27];
	x2 = p6 ^ rk[2];
	rk[3] ^= rk[28];
	x3 = p7 ^ rk[3];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[4] ^= rk[29];
	x0 ^= rk[4];
	rk[5] ^= rk[30];
	x1 ^= rk[5];
	rk[6] ^= rk[31];
	x2 ^= rk[6];
	rk[7] ^= rk[0];
	x3 ^= rk[7];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[8] ^= rk[1];
	x0 ^= rk[8];
	rk[9] ^= rk[2];
	x1 ^= rk[9];
	rk[10] ^= rk[3];
	x2 ^= rk[10];
	rk[11] ^= rk[4];
	x3 ^= rk[11];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[12] ^= rk[5];
	x0 ^= rk[12];
	rk[13] ^= rk[6];
	x1 ^= rk[13];
	rk[14] ^= rk[7];
	x2 ^= rk[14];
	rk[15] ^= rk[8];
	x3 ^= rk[15];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	p0 ^= x0;
	p1 ^= x1;
	p2 ^= x2;
	p3 ^= x3;
	rk[16] ^= rk[9];
	x0 = pC ^ rk[16];
	rk[17] ^= rk[10];
	x1 = pD ^ rk[17];
	rk[18] ^= rk[11];
	x2 = pE ^ rk[18];
	rk[19] ^= rk[12];
	x3 = pF ^ rk[19];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[20] ^= rk[13];
	x0 ^= rk[20];
	rk[21] ^= rk[14];
	x1 ^= rk[21];
	rk[22] ^= rk[15];
	x2 ^= rk[22];
	rk[23] ^= rk[16];
	x3 ^= rk[23];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[24] ^= rk[17];
	x0 ^= rk[24];
	rk[25] ^= rk[18];
	x1 ^= rk[25];
	rk[26] ^= rk[19];
	x2 ^= rk[26];
	rk[27] ^= rk[20];
	x3 ^= rk[27];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[28] ^= rk[21];
	x0 ^= rk[28];
	rk[29] ^= rk[22];
	x1 ^= rk[29];
	rk[30] ^= rk[23];
	x2 ^= rk[30];
	rk[31] ^= rk[24];
	x3 ^= rk[31];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	p8 ^= x0;
	p9 ^= x1;
	pA ^= x2;
	pB ^= x3;

	// 3
	KEY_EXPAND_ELT(sharedMemory, rk[0], rk[1], rk[2], rk[3]);
	rk[0] ^= rk[28];
	rk[1] ^= rk[29];
	rk[2] ^= rk[30];
	rk[3] ^= rk[31];
	x0 = p0 ^ rk[0];
	x1 = p1 ^ rk[1];
	x2 = p2 ^ rk[2];
	x3 = p3 ^ rk[3];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[4], rk[5], rk[6], rk[7]);
	rk[4] ^= rk[0];
	rk[5] ^= rk[1];
	rk[6] ^= rk[2];
	rk[7] ^= rk[3];
	x0 ^= rk[4];
	x1 ^= rk[5];
	x2 ^= rk[6];
	x3 ^= rk[7];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[8], rk[9], rk[10], rk[11]);
	rk[8] ^= rk[4];
	rk[9] ^= rk[5];
	rk[10] ^= rk[6];
	rk[11] ^= rk[7];
	x0 ^= rk[8];
	x1 ^= rk[9];
	x2 ^= rk[10];
	x3 ^= rk[11];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[12], rk[13], rk[14], rk[15]);
	rk[12] ^= rk[8];
	rk[13] ^= rk[9];
	rk[14] ^= rk[10];
	rk[15] ^= rk[11];
	x0 ^= rk[12];
	x1 ^= rk[13];
	x2 ^= rk[14];
	x3 ^= rk[15];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	pC ^= x0;
	pD ^= x1;
	pE ^= x2;
	pF ^= x3;
	KEY_EXPAND_ELT(sharedMemory, rk[16], rk[17], rk[18], rk[19]);
	rk[16] ^= rk[12];
	rk[17] ^= rk[13];
	rk[18] ^= rk[14];
	rk[19] ^= rk[15];
	x0 = p8 ^ rk[16];
	x1 = p9 ^ rk[17];
	x2 = pA ^ rk[18];
	x3 = pB ^ rk[19];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[20], rk[21], rk[22], rk[23]);
	rk[20] ^= rk[16];
	rk[21] ^= rk[17];
	rk[22] ^= rk[18];
	rk[23] ^= rk[19];
	x0 ^= rk[20];
	x1 ^= rk[21];
	x2 ^= rk[22];
	x3 ^= rk[23];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[24], rk[25], rk[26], rk[27]);
	rk[24] ^= rk[20];
	rk[25] ^= rk[21];
	rk[26] ^= rk[22];
	rk[27] ^= rk[23];
	x0 ^= rk[24];
	x1 ^= rk[25];
	x2 ^= rk[26];
	x3 ^= rk[27];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[28], rk[29], rk[30], rk[31]);
	rk[28] ^= rk[24];
	rk[29] ^= rk[25];
	rk[30] ^= rk[26];
	rk[31] ^= ~rk[27];
	rk[30] ^= counter;
	//rk[31] ^= 0xFFFFFFFF;
	x0 ^= rk[28];
	x1 ^= rk[29];
	x2 ^= rk[30];
	x3 ^= rk[31];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	p4 ^= x0;
	p5 ^= x1;
	p6 ^= x2;
	p7 ^= x3;

	rk[0] ^= rk[25];
	x0 = pC ^ rk[0];
	rk[1] ^= rk[26];
	x1 = pD ^ rk[1];
	rk[2] ^= rk[27];
	x2 = pE ^ rk[2];
	rk[3] ^= rk[28];
	x3 = pF ^ rk[3];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[4] ^= rk[29];
	x0 ^= rk[4];
	rk[5] ^= rk[30];
	x1 ^= rk[5];
	rk[6] ^= rk[31];
	x2 ^= rk[6];
	rk[7] ^= rk[0];
	x3 ^= rk[7];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[8] ^= rk[1];
	x0 ^= rk[8];
	rk[9] ^= rk[2];
	x1 ^= rk[9];
	rk[10] ^= rk[3];
	x2 ^= rk[10];
	rk[11] ^= rk[4];
	x3 ^= rk[11];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[12] ^= rk[5];
	x0 ^= rk[12];
	rk[13] ^= rk[6];
	x1 ^= rk[13];
	rk[14] ^= rk[7];
	x2 ^= rk[14];
	rk[15] ^= rk[8];
	x3 ^= rk[15];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	p8 ^= x0;
	p9 ^= x1;
	pA ^= x2;
	pB ^= x3;
	rk[16] ^= rk[9];
	x0 = p4 ^ rk[16];
	rk[17] ^= rk[10];
	x1 = p5 ^ rk[17];
	rk[18] ^= rk[11];
	x2 = p6 ^ rk[18];
	rk[19] ^= rk[12];
	x3 = p7 ^ rk[19];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[20] ^= rk[13];
	x0 ^= rk[20];
	rk[21] ^= rk[14];
	x1 ^= rk[21];
	rk[22] ^= rk[15];
	x2 ^= rk[22];
	rk[23] ^= rk[16];
	x3 ^= rk[23];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[24] ^= rk[17];
	x0 ^= rk[24];
	rk[25] ^= rk[18];
	x1 ^= rk[25];
	rk[26] ^= rk[19];
	x2 ^= rk[26];
	rk[27] ^= rk[20];
	x3 ^= rk[27];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[28] ^= rk[21];
	x0 ^= rk[28];
	rk[29] ^= rk[22];
	x1 ^= rk[29];
	rk[30] ^= rk[23];
	x2 ^= rk[30];
	rk[31] ^= rk[24];
	x3 ^= rk[31];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	p0 ^= x0;
	p1 ^= x1;
	p2 ^= x2;
	p3 ^= x3;


	/* round 3, 7, 11 */
	KEY_EXPAND_ELT(sharedMemory, rk[0], rk[1], rk[2], rk[3]);
	rk[0] ^= rk[28];
	rk[1] ^= rk[29];
	rk[2] ^= rk[30];
	rk[3] ^= rk[31];
	x0 = p8 ^ rk[0];
	x1 = p9 ^ rk[1];
	x2 = pA ^ rk[2];
	x3 = pB ^ rk[3];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[4], rk[5], rk[6], rk[7]);
	rk[4] ^= rk[0];
	rk[5] ^= rk[1];
	rk[6] ^= rk[2];
	rk[7] ^= rk[3];
	x0 ^= rk[4];
	x1 ^= rk[5];
	x2 ^= rk[6];
	x3 ^= rk[7];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[8], rk[9], rk[10], rk[11]);
	rk[8] ^= rk[4];
	rk[9] ^= rk[5];
	rk[10] ^= rk[6];
	rk[11] ^= rk[7];
	x0 ^= rk[8];
	x1 ^= rk[9];
	x2 ^= rk[10];
	x3 ^= rk[11];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[12], rk[13], rk[14], rk[15]);
	rk[12] ^= rk[8];
	rk[13] ^= rk[9];
	rk[14] ^= rk[10];
	rk[15] ^= rk[11];
	x0 ^= rk[12];
	x1 ^= rk[13];
	x2 ^= rk[14];
	x3 ^= rk[15];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	p4 ^= x0;
	p5 ^= x1;
	p6 ^= x2;
	p7 ^= x3;
	KEY_EXPAND_ELT(sharedMemory, rk[16], rk[17], rk[18], rk[19]);
	rk[16] ^= rk[12];
	rk[17] ^= rk[13];
	rk[18] ^= rk[14];
	rk[19] ^= rk[15];
	x0 = p0 ^ rk[16];
	x1 = p1 ^ rk[17];
	x2 = p2 ^ rk[18];
	x3 = p3 ^ rk[19];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[20], rk[21], rk[22], rk[23]);
	rk[20] ^= rk[16];
	rk[21] ^= rk[17];
	rk[22] ^= rk[18];
	rk[23] ^= rk[19];
	x0 ^= rk[20];
	x1 ^= rk[21];
	x2 ^= rk[22];
	x3 ^= rk[23];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[24], rk[25], rk[26], rk[27]);
	rk[24] ^= rk[20];
	rk[25] ^= rk[21];
	rk[26] ^= rk[22];
	rk[27] ^= rk[23];
	x0 ^= rk[24];
	x1 ^= rk[25];
	x2 ^= rk[26];
	x3 ^= rk[27];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[28], rk[29], rk[30], rk[31]);
	rk[28] ^= rk[24];
	rk[29] ^= rk[25];
	rk[30] ^= rk[26];
	rk[31] ^= rk[27];
	x0 ^= rk[28];
	x1 ^= rk[29];
	x2 ^= rk[30];
	x3 ^= rk[31];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	pC ^= x0;
	pD ^= x1;
	pE ^= x2;
	pF ^= x3;
	/* round 4, 8, 12 */
	rk[0] ^= rk[25];
	x0 = p4 ^ rk[0];
	rk[1] ^= rk[26];
	x1 = p5 ^ rk[1];
	rk[2] ^= rk[27];
	x2 = p6 ^ rk[2];
	rk[3] ^= rk[28];
	x3 = p7 ^ rk[3];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[4] ^= rk[29];
	x0 ^= rk[4];
	rk[5] ^= rk[30];
	x1 ^= rk[5];
	rk[6] ^= rk[31];
	x2 ^= rk[6];
	rk[7] ^= rk[0];
	x3 ^= rk[7];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[8] ^= rk[1];
	x0 ^= rk[8];
	rk[9] ^= rk[2];
	x1 ^= rk[9];
	rk[10] ^= rk[3];
	x2 ^= rk[10];
	rk[11] ^= rk[4];
	x3 ^= rk[11];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[12] ^= rk[5];
	x0 ^= rk[12];
	rk[13] ^= rk[6];
	x1 ^= rk[13];
	rk[14] ^= rk[7];
	x2 ^= rk[14];
	rk[15] ^= rk[8];
	x3 ^= rk[15];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	p0 ^= x0;
	p1 ^= x1;
	p2 ^= x2;
	p3 ^= x3;
	rk[16] ^= rk[9];
	x0 = pC ^ rk[16];
	rk[17] ^= rk[10];
	x1 = pD ^ rk[17];
	rk[18] ^= rk[11];
	x2 = pE ^ rk[18];
	rk[19] ^= rk[12];
	x3 = pF ^ rk[19];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[20] ^= rk[13];
	x0 ^= rk[20];
	rk[21] ^= rk[14];
	x1 ^= rk[21];
	rk[22] ^= rk[15];
	x2 ^= rk[22];
	rk[23] ^= rk[16];
	x3 ^= rk[23];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[24] ^= rk[17];
	x0 ^= rk[24];
	rk[25] ^= rk[18];
	x1 ^= rk[25];
	rk[26] ^= rk[19];
	x2 ^= rk[26];
	rk[27] ^= rk[20];
	x3 ^= rk[27];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	rk[28] ^= rk[21];
	x0 ^= rk[28];
	rk[29] ^= rk[22];
	x1 ^= rk[29];
	rk[30] ^= rk[23];
	x2 ^= rk[30];
	rk[31] ^= rk[24];
	x3 ^= rk[31];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	p8 ^= x0;
	p9 ^= x1;
	pA ^= x2;
	pB ^= x3;

	/* round 13 */
	KEY_EXPAND_ELT(sharedMemory, rk[0], rk[1], rk[2], rk[3]);
	rk[0] ^= rk[28];
	rk[1] ^= rk[29];
	rk[2] ^= rk[30];
	rk[3] ^= rk[31];
	x0 = p0 ^ rk[0];
	x1 = p1 ^ rk[1];
	x2 = p2 ^ rk[2];
	x3 = p3 ^ rk[3];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[4], rk[5], rk[6], rk[7]);
	rk[4] ^= rk[0];
	rk[5] ^= rk[1];
	rk[6] ^= rk[2];
	rk[7] ^= rk[3];
	x0 ^= rk[4];
	x1 ^= rk[5];
	x2 ^= rk[6];
	x3 ^= rk[7];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[8], rk[9], rk[10], rk[11]);
	rk[8] ^= rk[4];
	rk[9] ^= rk[5];
	rk[10] ^= rk[6];
	rk[11] ^= rk[7];
	x0 ^= rk[8];
	x1 ^= rk[9];
	x2 ^= rk[10];
	x3 ^= rk[11];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[12], rk[13], rk[14], rk[15]);
	rk[12] ^= rk[8];
	rk[13] ^= rk[9];
	rk[14] ^= rk[10];
	rk[15] ^= rk[11];
	x0 ^= rk[12];
	x1 ^= rk[13];
	x2 ^= rk[14];
	x3 ^= rk[15];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	pC ^= x0;
	pD ^= x1;
	pE ^= x2;
	pF ^= x3;
	KEY_EXPAND_ELT(sharedMemory, rk[16], rk[17], rk[18], rk[19]);
	rk[16] ^= rk[12];
	rk[17] ^= rk[13];
	rk[18] ^= rk[14];
	rk[19] ^= rk[15];
	x0 = p8 ^ rk[16];
	x1 = p9 ^ rk[17];
	x2 = pA ^ rk[18];
	x3 = pB ^ rk[19];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[20], rk[21], rk[22], rk[23]);
	rk[20] ^= rk[16];
	rk[21] ^= rk[17];
	rk[22] ^= rk[18];
	rk[23] ^= rk[19];
	x0 ^= rk[20];
	x1 ^= rk[21];
	x2 ^= rk[22];
	x3 ^= rk[23];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[24], rk[25], rk[26], rk[27]);
	rk[24] ^= rk[20];
	rk[25] ^= rk[21] ^ counter;
	rk[26] ^= rk[22];
	rk[27] ^= ~rk[23]; //^ 0xFFFFFFFF;
	x0 ^= rk[24];
	x1 ^= rk[25];
	x2 ^= rk[26];
	x3 ^= rk[27];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	KEY_EXPAND_ELT(sharedMemory, rk[28], rk[29], rk[30], rk[31]);
	rk[28] ^= rk[24];
	rk[29] ^= rk[25];
	rk[30] ^= rk[26];
	rk[31] ^= rk[27];
	x0 ^= rk[28];
	x1 ^= rk[29];
	x2 ^= rk[30];
	x3 ^= rk[31];
	AES_ROUND_NOKEY(sharedMemory, x0, x1, x2, x3);
	p4 ^= x0;
	p5 ^= x1;
	p6 ^= x2;
	p7 ^= x3;
	state[0x0] ^= p8;
	state[0x1] ^= p9;
	state[0x2] ^= pA;
	state[0x3] ^= pB;
	state[0x4] ^= pC;
	state[0x5] ^= pD;
	state[0x6] ^= pE;
	state[0x7] ^= pF;
	state[0x8] ^= p0;
	state[0x9] ^= p1;
	state[0xA] ^= p2;
	state[0xB] ^= p3;
	state[0xC] ^= p4;
	state[0xD] ^= p5;
	state[0xE] ^= p6;
	state[0xF] ^= p7;
}

__global__ __launch_bounds__(TPB, 3)
void x11_shavite512_gpu_hash_80(uint32_t threads, uint32_t startNounce, void *outputHash)
{
	__shared__ uint32_t sharedMemory[1024];

	if(threadIdx.x < 256)
	{
		sharedMemory[threadIdx.x] = d_AES0[threadIdx.x];
		sharedMemory[threadIdx.x + 256] = ROTL32(sharedMemory[threadIdx.x], 8);
		sharedMemory[threadIdx.x + 512] = ROTL32(sharedMemory[threadIdx.x], 16);
		sharedMemory[threadIdx.x + 768] = ROTL32(sharedMemory[threadIdx.x], 24);
	}
	__syncthreads();
	const uint32_t thread = (blockDim.x * blockIdx.x + threadIdx.x);
	if(thread < threads)
	{
		const uint32_t nounce = startNounce + thread;

		// kopiere init-state
		uint32_t state[16] = {
			0x72FCCDD8, 0x79CA4727, 0x128A077B, 0x40D55AEC,
			0xD1901A06, 0x430AE307, 0xB29F5CD1, 0xDF07FBFC,
			0x8E45D73D, 0x681AB538, 0xBDE86578, 0xDD577E47,
			0xE275EADE, 0x502D9FCD, 0xB9357178, 0x022A4B9A
		};

		uint32_t msg[32];

#pragma unroll
		for(int i = 0; i<31; i++)
		{
			msg[i] = c_PaddedMessage80[i];
		}
		msg[19] = cuda_swab32(nounce);
		msg[20] = 0x80;
		msg[27] = 0x2800000;
		msg[31] = 0x2000000;

		c512(sharedMemory, state, msg);

		uint32_t *outHash = (uint32_t *)outputHash + 16 * thread;

#pragma unroll 16
		for(int i = 0; i<16; i++)
			outHash[i] = state[i];

	} //thread < threads
}

__host__ void x11_shavite512_cpu_hash_64(int thr_id, uint32_t threads, uint32_t startNounce, uint32_t *d_hash)
{
	// berechne wie viele Thread Blocks wir brauchen
	dim3 grid((threads + TPB-1)/TPB);
	dim3 block(TPB);

	x11_shavite512_gpu_hash_64<<<grid, block, 0, gpustream[thr_id]>>>(threads, d_hash);

	CUDA_SAFE_CALL(cudaGetLastError());
}

__host__ void x11_shavite512_cpu_hash_80(int thr_id, uint32_t threads, uint32_t startNounce, uint32_t *d_outputHash)
{

	// berechne wie viele Thread Blocks wir brauchen
	dim3 grid((threads + TPB-1)/TPB);
	dim3 block(TPB);

	x11_shavite512_gpu_hash_80<<<grid, block, 0, gpustream[thr_id]>>>(threads, startNounce, d_outputHash);
}

__host__ void x11_shavite512_setBlock_80(int thr_id, void *pdata)
{
	// Message mit Padding bereitstellen
	// lediglich die korrekte Nonce ist noch ab Byte 76 einzusetzen.
	unsigned char PaddedMessage[128];
	memcpy(PaddedMessage, pdata, 80);
	memset(PaddedMessage+80, 0, 48);

	cudaMemcpyToSymbolAsync(c_PaddedMessage80, PaddedMessage, 32 * sizeof(uint32_t), 0, cudaMemcpyHostToDevice, gpustream[thr_id]);
}

